use strict;
use warnings;

package Net::Bugzilla::Kanbanize::Bug;
$Net::Bugzilla::Kanbanize::Bug::VERSION = '0.007';    # TRIAL
1;

__END__

=pod

=encoding UTF-8

=head1 NAME

Net::Bugzilla::Kanbanize::Bug

=head1 VERSION

version 0.007

=head1 SYNOPSIS

Kanbanize Bugzilla Bug Abstraction

=head1 AUTHOR

Philippe M. Chiasson <gozer@cpan.org>

=head1 COPYRIGHT AND LICENSE

This software is Copyright (c) 2014 by Philippe M. Chiasson.

This is free software, licensed under:

  Mozilla Public License Version 2.0

=cut
